import torch.nn as nn
import torch
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
import math
from modules import ir_1w1a, straightThrough, gumbel_quant, birealnetModules, gradientEstimators

# CUDA_VISIBLE_DEVICES=7 python trainSimple.py --arch bnnA1BK3 --save-dir /media/data4b/haleyso/bnn/large/A1BK3 --method ste

__all__ = ['blitz0', "blitzW1A1", "blitzW1", "blitzGumbelSoftmax", "blitzGumbelSoftmaxW1A1", "blitzSTEW1A1", "blitzSTEW1", "blitzBiRealNet",
           "blitzgclipW1", "blitzgclipW1A1", "blitzgSecondOW1", "blitzgSecondOW1A1", "blitzgBCW1", "blitzgBCW1A1", "blitzgirW1","blitzgirW1A1",
           "bnnA5BIR","bnnA5B","bnnA10B" , "bnnA50B", "bnnA1BK3" , "bnnABK3", "bnnF256","bnnF256gumbel", "vgg_small_1w1a_gumbel", "vgg_small_1w_gumbel"]



class BLITZ(nn.Module):
# full precision
    def __init__(self, num_classes=10):
        super(BLITZ, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        # self.convhalf = nn.Conv2d(6, 6, 5)
        self.conv2 = nn.Conv2d(6, 16, 5)

        bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=bias)
        self.fc2 = nn.Linear(120, 84, bias=bias)
        self.fc3 = nn.Linear(84, 10, bias=bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZW1A1(nn.Module):
# using the ir-net method
    def __init__(self, num_classes=10):
        super(BLITZW1A1, self).__init__()

        # self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        # self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        self.conv1 = ir_1w1a.IRConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = ir_1w1a.IRConv2d(in_channels=6, out_channels=16, kernel_size=5)
        bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=bias)
        self.fc2 = nn.Linear(120, 84, bias=bias)
        self.fc3 = nn.Linear(84, 10, bias=bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BLITZW1(nn.Module):
# using the ir-net method
    def __init__(self, num_classes=10):
        super(BLITZW1, self).__init__()

        # self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        # self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        self.conv1 = ir_1w1a.IRConv2dW1(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = ir_1w1a.IRConv2dW1(in_channels=6, out_channels=16, kernel_size=5)
        bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=bias)
        self.fc2 = nn.Linear(120, 84, bias=bias)
        self.fc3 = nn.Linear(84, 10, bias=bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZSTEW1A1(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BLITZSTEW1A1, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        # self.conv1 = ir_1w1a.IRConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        # self.conv2 = ir_1w1a.IRConv2d(in_channels=6, out_channels=16, kernel_size=5)
        self.bias = True
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BLITZSTEW1(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BLITZSTEW1, self).__init__()

        self.conv1 = straightThrough.STEConv2d_noA(in_channels=3, out_channels=6, kernel_size=5)
        # self.conv1 = ir_1w1a.IRConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = straightThrough.STEConv2d_noA(in_channels=6, out_channels=16,kernel_size=5)
        # self.conv2 = ir_1w1a.IRConv2d(in_channels=6, out_channels=16, kernel_size=5)
        self.bias = True
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BLITZGUMBEL(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGUMBEL, self).__init__()
        self.conv1 = gumbel_quant.GumbelConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gumbel_quant.GumbelConv2d(in_channels=6, out_channels=16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x, self.conv1.tau, hard=True)))
        x = self.pool(F.relu(self.conv2(x, self.conv2.tau, hard=True)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BLITZGUMBELW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGUMBELW1A1, self).__init__()
        self.conv1 = gumbel_quant.GumbelConv2dW1A1(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gumbel_quant.GumbelConv2dW1A1(in_channels=6, out_channels=16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x, self.conv1.tau, hard=True)))
        x = self.pool(F.relu(self.conv2(x, self.conv2.tau, hard=True)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BLITZBIREALW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZBIREALW1A1, self).__init__()
        self.binary_activation1 = birealnetModules.BinaryActivation()
        self.conv1 = birealnetModules.HardBinaryConv(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.binary_activation2 = birealnetModules.BinaryActivation()
        self.conv2 = birealnetModules.HardBinaryConv(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

#------------------------------------------------------------------------------------------------------------------------------
class BLITZGCLIPW1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGCLIPW1, self).__init__()
        self.conv1 = gradientEstimators.clipConv2dW1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.clipConv2dW1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGCLIPW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGCLIPW1A1, self).__init__()
        self.conv1 = gradientEstimators.clipConv2dW1A1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.clipConv2dW1A1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x



class BLITZGSECONDOW1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGSECONDOW1, self).__init__()
        self.conv1 = gradientEstimators.secondOrderConv2dW1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.secondOrderConv2dW1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGSECONDOW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGSECONDOW1A1, self).__init__()
        self.conv1 = gradientEstimators.secondOrderConv2dW1A1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.secondOrderConv2dW1A1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)


    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGBCW1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGBCW1, self).__init__()
        self.conv1 = gradientEstimators.bcConv2dW1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.bcConv2dW1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)


    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGBCW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGBCW1A1, self).__init__()
        self.conv1 = gradientEstimators.bcConv2dW1A1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.bcConv2dW1A1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)


    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGIRW1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGIRW1, self).__init__()
        self.conv1 = gradientEstimators.irConv2dW1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.irConv2dW1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)


    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BLITZGIRW1A1(nn.Module):
    def __init__(self, num_classes=10):
        super(BLITZGIRW1A1, self).__init__()
        self.conv1 = gradientEstimators.irConv2dW1A1(3, 6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gradientEstimators.irConv2dW1A1(6, 16, kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, self.bias)
        self.fc2 = nn.Linear(120, 84, self.bias)
        self.fc3 = nn.Linear(84, 10, self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


#--------------------------LARGE----------------
class BNNA5BIR(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNA5BIR, self).__init__()

        self.conv1 = gradientEstimators.irConv2dW1(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conva = gradientEstimators.irConv2dW1(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb = gradientEstimators.irConv2dW1(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc = gradientEstimators.irConv2dW1(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd = gradientEstimators.irConv2dW1(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve = gradientEstimators.irConv2dW1(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conv2 = gradientEstimators.irConv2dW1(in_channels=6, out_channels=16,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = F.relu(self.conva(x))
        x = F.relu(self.convb(x))
        x = F.relu(self.convc(x))
        x = F.relu(self.convd(x))
        x = F.relu(self.conve(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BNNA5B(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNA5B, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conva = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = F.relu(self.conva(x))
        x = F.relu(self.convb(x))
        x = F.relu(self.convc(x))
        x = F.relu(self.convd(x))
        x = F.relu(self.conve(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BNNA10B(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNA10B, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conva = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = F.relu(self.conva(x))
        x = F.relu(self.convb(x))
        x = F.relu(self.convc(x))
        x = F.relu(self.convd(x))
        x = F.relu(self.conve(x))
        x = F.relu(self.convf(x))
        x = F.relu(self.convg(x))
        x = F.relu(self.convh(x))
        x = F.relu(self.convi(x))
        x = F.relu(self.convj(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BNNA50B(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNA50B, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conva = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conva2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj2 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conva3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj3 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conva4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj4 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conva5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convb5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convc5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convd5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.conve5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convf5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convg5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convh5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convi5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)
        self.convj5 = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=5, padding=2)

        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(16 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = F.relu(self.conva(x))
        x = F.relu(self.convb(x))
        x = F.relu(self.convc(x))
        x = F.relu(self.convd(x))
        x = F.relu(self.conve(x))
        x = F.relu(self.convf(x))
        x = F.relu(self.convg(x))
        x = F.relu(self.convh(x))
        x = F.relu(self.convi(x))
        x = F.relu(self.convj(x))

        x = F.relu(self.conva2(x))
        x = F.relu(self.convb2(x))
        x = F.relu(self.convc2(x))
        x = F.relu(self.convd2(x))
        x = F.relu(self.conve2(x))
        x = F.relu(self.convf2(x))
        x = F.relu(self.convg2(x))
        x = F.relu(self.convh2(x))
        x = F.relu(self.convi2(x))
        x = F.relu(self.convj2(x))

        x = F.relu(self.conva3(x))
        x = F.relu(self.convb3(x))
        x = F.relu(self.convc3(x))
        x = F.relu(self.convd3(x))
        x = F.relu(self.conve3(x))
        x = F.relu(self.convf3(x))
        x = F.relu(self.convg3(x))
        x = F.relu(self.convh3(x))
        x = F.relu(self.convi3(x))
        x = F.relu(self.convj3(x))

        x = F.relu(self.conva4(x))
        x = F.relu(self.convb4(x))
        x = F.relu(self.convc4(x))
        x = F.relu(self.convd4(x))
        x = F.relu(self.conve4(x))
        x = F.relu(self.convf4(x))
        x = F.relu(self.convg4(x))
        x = F.relu(self.convh4(x))
        x = F.relu(self.convi4(x))
        x = F.relu(self.convj4(x))

        x = F.relu(self.conva5(x))
        x = F.relu(self.convb5(x))
        x = F.relu(self.convc5(x))
        x = F.relu(self.convd5(x))
        x = F.relu(self.conve5(x))
        x = F.relu(self.convf5(x))
        x = F.relu(self.convg5(x))
        x = F.relu(self.convh5(x))
        x = F.relu(self.convi5(x))
        x = F.relu(self.convj5(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BNNA1BK3(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNA1BK3, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=3)
        self.pool = nn.MaxPool2d(2, 2)
        self.conva = straightThrough.STEConv2d(in_channels=6, out_channels=6, kernel_size=3, padding=2)

        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=3)
        self.bias = False
        self.fc1 = nn.Linear(16 * 7 * 7, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = F.relu(self.conva(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class BNNABK3(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNABK3, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=6, kernel_size=3)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = straightThrough.STEConv2d(in_channels=6, out_channels=16,kernel_size=3)
        self.bias = False
        self.fc1 = nn.Linear(16 * 6 * 6, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        # x = F.relu(self.conva(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BNNF256(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNF256, self).__init__()

        self.conv1 = straightThrough.STEConv2d(in_channels=3, out_channels=64, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = straightThrough.STEConv2d(in_channels=64, out_channels=256,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(256 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        # x = F.relu(self.conva(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class BNNF256gumbel(nn.Module):
# using the ir-net or ste estimator and training methods
    def __init__(self, num_classes=10):
        super(BNNF256gumbel, self).__init__()

        self.conv1 = gumbel_quant.GumbelConv2d(in_channels=3, out_channels=64, kernel_size=5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = gumbel_quant.GumbelConv2d(in_channels=64, out_channels=256,kernel_size=5)
        self.bias = False
        self.fc1 = nn.Linear(256 * 5 * 5, 120, bias=self.bias)
        self.fc2 = nn.Linear(120, 84, bias=self.bias)
        self.fc3 = nn.Linear(84, 10, bias=self.bias)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        # x = F.relu(self.conva(x))
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = torch.flatten(x, 1)     # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class VGG_SMALL_1W1A_GUMBEL(nn.Module):

    def __init__(self, num_classes=10):
        super(VGG_SMALL_1W1A_GUMBEL, self).__init__()
        self.conv0 = nn.Conv2d(3, 128, kernel_size=3, padding=1, bias=False)
        self.bn0 = nn.BatchNorm2d(128)
        self.conv1 = gumbel_quant.GumbelConv2dW1A1(128, 128, kernel_size=3, padding=1, bias=False)
        self.pooling = nn.MaxPool2d(kernel_size=2, stride=2)
        self.bn1 = nn.BatchNorm2d(128)
        # self.nonlinear = nn.ReLU(inplace=True)
        self.nonlinear = nn.Hardtanh(inplace=True)
        self.conv2 = gumbel_quant.GumbelConv2dW1A1(128, 256, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(256)
        self.conv3 = gumbel_quant.GumbelConv2dW1A1(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(256)
        self.conv4 = gumbel_quant.GumbelConv2dW1A1(256, 512, kernel_size=3, padding=1, bias=False)
        self.bn4 = nn.BatchNorm2d(512)
        self.conv5 = gumbel_quant.GumbelConv2dW1A1(512, 512, kernel_size=3, padding=1, bias=False)
        self.bn5 = nn.BatchNorm2d(512)
        self.fc = nn.Linear(512*4*4, num_classes)
        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, gumbel_quant.GumbelConv2dW1A1):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()

    def forward(self, x):
        x = self.conv0(x)
        x = self.bn0(x)
        x = self.nonlinear(x)
        x = self.conv1(x)
        x = self.pooling(x)
        x = self.bn1(x)
        x = self.nonlinear(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.nonlinear(x)
        x = self.conv3(x)
        x = self.pooling(x)
        x = self.bn3(x)
        x = self.nonlinear(x)
        x = self.conv4(x)
        x = self.bn4(x)
        x = self.nonlinear(x)
        x = self.conv5(x)
        x = self.pooling(x)
        x = self.bn5(x)
        x = self.nonlinear(x)
        # x = self.pooling(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x

class VGG_SMALL_1W_GUMBEL(nn.Module):

    def __init__(self, num_classes=10):
        super(VGG_SMALL_1W_GUMBEL, self).__init__()
        self.conv0 = nn.Conv2d(3, 128, kernel_size=3, padding=1, bias=False)
        self.bn0 = nn.BatchNorm2d(128)
        self.conv1 = gumbel_quant.GumbelConv2d(128, 128, kernel_size=3, padding=1, bias=False)
        self.pooling = nn.MaxPool2d(kernel_size=2, stride=2)
        self.bn1 = nn.BatchNorm2d(128)
        # self.nonlinear = nn.ReLU(inplace=True)
        self.nonlinear = nn.Hardtanh(inplace=True)
        self.conv2 = gumbel_quant.GumbelConv2d(128, 256, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(256)
        self.conv3 = gumbel_quant.GumbelConv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(256)
        self.conv4 = gumbel_quant.GumbelConv2d(256, 512, kernel_size=3, padding=1, bias=False)
        self.bn4 = nn.BatchNorm2d(512)
        self.conv5 = gumbel_quant.GumbelConv2d(512, 512, kernel_size=3, padding=1, bias=False)
        self.bn5 = nn.BatchNorm2d(512)
        self.fc = nn.Linear(512*4*4, num_classes)
        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, gumbel_quant.GumbelConv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()

    def forward(self, x):
        x = self.conv0(x)
        x = self.bn0(x)
        x = self.nonlinear(x)
        x = self.conv1(x)
        x = self.pooling(x)
        x = self.bn1(x)
        x = self.nonlinear(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.nonlinear(x)
        x = self.conv3(x)
        x = self.pooling(x)
        x = self.bn3(x)
        x = self.nonlinear(x)
        x = self.conv4(x)
        x = self.bn4(x)
        x = self.nonlinear(x)
        x = self.conv5(x)
        x = self.pooling(x)
        x = self.bn5(x)
        x = self.nonlinear(x)
        # x = self.pooling(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x

def blitz0(**kwargs):
    model = BLITZ(**kwargs)
    return model


def blitzW1A1(**kwargs):
    model = BLITZW1A1(**kwargs)
    return model

def blitzW1(**kwargs):
    model = BLITZW1(**kwargs)
    return model


def blitzGumbelSoftmax(**kwargs):
    model = BLITZGUMBEL(**kwargs)
    return model

def blitzGumbelSoftmaxW1A1(**kwargs):
    model =BLITZGUMBELW1A1(**kwargs)
    return model

def blitzSTEW1A1(**kwargs):
    model = BLITZSTEW1A1(**kwargs)
    return model


def blitzSTEW1(**kwargs):
    model = BLITZSTEW1(**kwargs)
    return model

def blitzBiRealNet(**kwargs):
    model = BLITZBIREALW1A1(**kwargs)
    return model
#---------------------------------------------
def blitzgclipW1(**kwargs):
    model = BLITZGCLIPW1(**kwargs)
    return model

def blitzgclipW1A1(**kwargs):
    model = BLITZGCLIPW1A1(**kwargs)
    return model

def blitzgSecondOW1(**kwargs):
    model = BLITZGSECONDOW1(**kwargs)
    return model

def blitzgSecondOW1A1(**kwargs):
    model = BLITZGSECONDOW1A1(**kwargs)
    return model

def blitzgBCW1(**kwargs):
    model = BLITZGBCW1(**kwargs)
    return model

def blitzgBCW1A1(**kwargs):
    model = BLITZGBCW1A1(**kwargs)
    return model

def blitzgirW1(**kwargs):
    model = BLITZGIRW1(**kwargs)
    return model

def blitzgirW1A1(**kwargs):
    model = BLITZGIRW1A1(**kwargs)
    return model

def bnnA5BIR(**kwargs):
    model = BNNA5BIR(**kwargs)
    return model

def bnnA5B(**kwargs):
    model = BNNA5B(**kwargs)
    return model

def bnnA10B(**kwargs):
    model = BNNA10B(**kwargs)
    return model

def bnnA50B(**kwargs):
    model = BNNA50B(**kwargs)
    return model

def bnnA1BK3(**kwargs):
    model = BNNA1BK3(**kwargs)
    return model

def bnnABK3(**kwargs):
    model = BNNABK3(**kwargs)
    return model

def bnnF256(**kwargs):
    model = BNNF256(**kwargs)
    return model

def bnnF256gumbel(**kwargs):
    model = BNNF256gumbel(**kwargs)
    return model


def vgg_small_1w1a_gumbel(**kwargs):
    model = VGG_SMALL_1W1A_GUMBEL(**kwargs)
    return model

def vgg_small_1w_gumbel(**kwargs):
    model = VGG_SMALL_1W_GUMBEL(**kwargs)
    return model
