import math
import sys

import torch
import torch.nn as nn
import torch.nn.functional as F


def score_quant_value(realWeight, s=5., func='sigmoid'):
    # Here s is kinda representing the steepness
    # score of how close a value is to the quantized values
    diff = torch.stack([realWeight + torch.ones_like(realWeight), realWeight - torch.ones_like(realWeight)]).permute(1,2,3,4,0).cuda()
    z = s * diff
    scores = torch.sigmoid(z) * (1 - torch.sigmoid(z)) * 4
    return scores


class GumbelConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(GumbelConv2d, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)
        self.tau = 0.1
        self.c = 300  # boost the score
        self.tau_max = 3.0
        self.q_weights = 0

    def forward(self, input, tau=0.2, hard=True):
        w = self.weight
        scores = score_quant_value(w, (self.tau_max / tau)**1) * self.c * (self.tau_max / tau)**0.5
        one_hot = F.gumbel_softmax(scores, tau=tau, hard=hard, dim=1)
        q_weights = (one_hot * torch.tensor([-1,1]).cuda())   # value of one hot vector  dot product with lut [-1,1]
        q_weights = q_weights.sum(dim=q_weights.ndim-1)
        a = input
        self.q_weights = q_weights
        output = F.conv2d(a, q_weights, self.bias, self.stride, self.padding,self.dilation, self.groups)
        return output


class GumbelConv2dW1A1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(GumbelConv2dW1A1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)
        self.tau = 0.1
        self.c = 300  # boost the score
        self.tau_max = 3.0
        self.q_weights = 0
        self.q_activations = 0

    def forward(self, input, tau=0.2, hard=True):
        w = self.weight
        scores = score_quant_value(w, (self.tau_max / tau)**1) * self.c * (self.tau_max / tau)**0.5
        one_hot = F.gumbel_softmax(scores, tau=tau, hard=hard, dim=1)
        q_weights = (one_hot * torch.tensor([-1,1]).cuda())   # value of one hot vector  dot product with lut [-1,1]
        q_weights = q_weights.sum(dim=q_weights.ndim-1)
        self.q_weights = q_weights

        a = input
        scores_a = score_quant_value(a, (self.tau_max / tau)**1) * self.c * (self.tau_max / tau)**0.5
        one_hot_a = F.gumbel_softmax(scores_a, tau=tau, hard=hard, dim=1)
        q_activations = (one_hot_a * torch.tensor([-1, 1]).cuda())
        q_activations = q_activations.sum(dim=q_activations.ndim - 1)
        self.q_activations = q_activations

        output = F.conv2d(q_activations, q_weights, self.bias, self.stride, self.padding,self.dilation, self.groups)
        return output

# class SoftmaxBasedQuantization(nn.Module):      # changes forward
#     def __init__(self, gumbel=True, tau_max=3.0, c=300.):
#         super(SoftmaxBasedQuantization, self).__init__()
#
#         lut = [-1, 1]
#         self.lut = torch.tensor(lut, dtype=torch.float32).reshape(1, len(lut), 1, 1)    # list of discrete values our weights can be
#         self.c = c  # boost the score
#         self.tau_max = tau_max
#
#     def forward(self, real_value, tau=0.1, hard=False):
#         scores = score_quant_value(real_value, self.lut.to(real_value.device), (self.tau_max / tau)**1) * self.c * (self.tau_max / tau)**0.5
#         # score to one-hot encoding
#         # (N, 1, H, W) -> (N, C, H, W)
#         one_hot = F.gumbel_softmax(scores, tau=tau, hard=hard, dim=1)
#
#         # one-hot encoding to phase value
#         q_value = (one_hot * self.lut.to(one_hot.device))   # value of one hot vector  dot product with lut [-1,1]
#         q_value = q_value.sum(1, keepdims=True)
#         return q_value



