import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
import torch
import math


class STEQuantize(Function):
    @staticmethod
    def forward(ctx, input):
        out = torch.sign(input)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        # output = grad_output
        # return Function.hardtanh(grad_output)
        return grad_output


class STEConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(STEConv2d, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        a = input
        # bw = w - w.view(w.size(0), -1).mean(-1).view(w.size(0), 1, 1, 1)
        # bw = bw / bw.view(bw.size(0), -1).std(-1).view(bw.size(0), 1, 1, 1)
        # sw = torch.pow(torch.tensor([2]*bw.size(0)).cuda().float(), (torch.log(bw.abs().view(bw.size(0), -1).mean(-1)) / math.log(2)).round().float()).view(bw.size(0), 1, 1, 1).detach()
        bw = STEQuantize().apply(w)
        ba = STEQuantize().apply(a)

        # bw = bw * sw
        output = F.conv2d(ba, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output

class STEConv2d_noA(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(STEConv2d_noA, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        a = input
        # bw = w - w.view(w.size(0), -1).mean(-1).view(w.size(0), 1, 1, 1)
        # bw = bw / bw.view(bw.size(0), -1).std(-1).view(bw.size(0), 1, 1, 1)
        # sw = torch.pow(torch.tensor([2]*bw.size(0)).cuda().float(), (torch.log(bw.abs().view(bw.size(0), -1).mean(-1)) / math.log(2)).round().float()).view(bw.size(0), 1, 1, 1).detach()
        bw = STEQuantize().apply(w)
        # ba = STEQuantize().apply(a)

        # bw = bw * sw
        output = F.conv2d(a, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output



#
# class STELinear(nn.Linear):
#     def __init__(self, in_channels, out_channels, bias=True):
#         super(STELinear, self).__init__(in_channels, out_channels, bias)
#         self.haley = 0
#
#     def forward(self, input):
#         w = self.weight
#         a = input
#         bw = w - w.view(w.size(0), -1).mean(-1).view(w.size(0), 1, 1, 1)
#         bw = bw / bw.view(bw.size(0), -1).std(-1).view(bw.size(0), 1, 1, 1)
#         sw = torch.pow(torch.tensor([2]*bw.size(0)).cuda().float(), (torch.log(bw.abs().view(bw.size(0), -1).mean(-1)) / math.log(2)).round().float()).view(bw.size(0), 1, 1, 1).detach()
#         bw = ste.STEQuantize().apply(bw)
#         ba = ste.STEQuantize().apply(a)
#
#         bw = bw * sw
#         output = F.linear(ba, bw, self.bias)
#         return output